
#ifdef USE_MPI
#include "common/mpi_helpers.h"
#endif /* USE_MPI */

#include "maxwell/maxwell_postpro.h"

namespace maxwell {

field_values
eval_field(const field& f, size_t ofs, size_t cbs, const vecxd& phi)
{
    assert(ofs+cbs <= f.num_dofs);
    field_values ret;
    ret.Ex = f.Ex.segment(ofs, cbs).dot(phi);
    ret.Ey = f.Ey.segment(ofs, cbs).dot(phi);
    ret.Ez = f.Ez.segment(ofs, cbs).dot(phi);
    ret.Hx = f.Hx.segment(ofs, cbs).dot(phi);
    ret.Hy = f.Hy.segment(ofs, cbs).dot(phi);
    ret.Hz = f.Hz.segment(ofs, cbs).dot(phi);
    return ret;
}

field_values
compute_error(const model& mod, const solver_state& state, const parameter_loader& mpl)
{
    field_values err;

    if (not mpl.has_analytical_solution())
        throw std::invalid_argument("analytical solution not defined in the Lua configuration");

    for (auto& e : mod)
    {
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            const auto& pe = e.cell(iT);
            const auto& re = e.cell_refelem(pe);

            const auto pqps = pe.integration_points();
            const auto dets = pe.determinants();

            const auto ofs = e.cell_model_dof_offset(iT);
            const auto cbs = re.num_basis_functions();

            assert(pqps.size() == dets.size());
            for (size_t iQp = 0; iQp < pqps.size(); iQp++)
            {
                const auto& pqp = pqps[iQp];
                const vecxd phi = re.basis_functions(iQp);

                field_values Fh = eval_field(state.emf_curr, ofs, cbs, phi);

                auto [E, H] = mpl.eval_analytical_solution(e.material_tag(),
                    pqp.point(), state.curr_time);
                field_values F(E,H);
                field_values Fdiff = F - Fh;
                err += Fdiff*Fdiff*pqp.weight();
            }
        }
    }

#ifdef USE_MPI
    MPI_Allreduce(MPI_IN_PLACE, &err, 6, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif /* USE_MPI */
    return sqrt(err);
}

field_values
compute_energy(const model& mod, const solver_state& state, const parameter_loader& mpl)
{
    field_values energy;

    for (auto& e : mod)
    {
        for (size_t iT = 0; iT < e.num_cells(); iT++)
        {
            const auto& pe = e.cell(iT);
            const auto& re = e.cell_refelem(pe);

            const auto pqps = pe.integration_points();
            const auto dets = pe.determinants();

            const auto ofs = e.cell_model_dof_offset(iT);
            const auto cbs = re.num_basis_functions();

            assert(pqps.size() == dets.size());
            const auto bar = pe.barycenter();
            auto eps = mpl.epsilon( e.material_tag(), bar );
            auto mu = mpl.mu( e.material_tag(), bar );

            field_values mat(eps, eps, eps, mu, mu, mu);

            for (size_t iQp = 0; iQp < pqps.size(); iQp++)
            {
                const auto& pqp = pqps[iQp];
                const vecxd phi = re.basis_functions(iQp);
                field_values Fh = eval_field(state.emf_curr, ofs, cbs, phi);
                energy += 0.5*mat*Fh*Fh*pqp.weight();
            }
        }
    }

#ifdef USE_MPI
    MPI_Allreduce(MPI_IN_PLACE, &energy, 6, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif /* USE_MPI */
    return energy;
}

#ifdef ENABLE_GPU_SOLVER
static void
validate(const model&, maxwell::solver_state_gpu&,
    maxwell::parameter_loader&, std::ofstream&)
{}
#endif /* ENABLE_GPU_SOLVER */

} // namespace maxwell

